from robin_stocks.gemini.authentication import generate_signature
from robin_stocks.gemini.helper import (format_inputs, login_required,
                                        request_post)
from robin_stocks.gemini.urls import URLS


@login_required
@format_inputs
def get_account_detail(jsonify=None):
    """ Gets information about the profile attached to your API key.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * account - Contains information on the requested account
                        -- accountName - The name of the account provided upon creation. Will default to Primary
                        -- shortName - Nickname of the specific account (will take the name given, remove all symbols, replace all " " with "-" and make letters lowercase)
                        -- type - The type of account. Will return either exchange or custody
                        -- created - The timestamp of account creation, displayed as number of milliseconds since 1970-01-01 UTC. This will be transmitted as a JSON number
                      * users - Contains an array of JSON objects with user information for the requested account
                        -- name - Full legal name of the user
                        -- lastSignIn - Timestamp of the last sign for the user. Formatted as yyyy-MM-dd'T'HH:mm:ss.SSS'Z'
                        -- status - Returns user status. Will inform of active users or otherwise not active
                        -- countryCode - 2 Letter country code indicating residence of user.
                        -- isVerified - Returns verification status of user.
                      * memo_reference_code - Returns wire memo reference code for linked bank account.

    """
    url = URLS.account_detail()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def check_available_balances(jsonify=None):
    """ Gets a list of all available balances in every currency.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * currency - The currency code.
                      * amount - The current balance
                      * available - The amount that is available to trade
                      * availableForWithdrawal - The amount that is available to withdraw
                      * type - "exchange"

    """
    url = URLS.available_balances()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def check_notional_balances(jsonify=None):
    """ Gets a list of all available balances in every currency.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.\
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * currency - The currency code.
                      * amount - The current balance
                      * amountNotional - Amount, in notional
                      * available - The amount that is available to trade
                      * availableNotional - Available, in notional
                      * availableForWithdrawal - The amount that is available to withdraw
                      * availableForWithdrawalNotional - AvailableForWithdrawal, in notional

    """
    url = URLS.notional_balances()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def check_transfers(timestamp=None, limit_transfers=10, show_completed_deposit_advances=False, jsonify=None):
    """ Gets a list of all transfers.

    :param timestamp: Only return transfers on or after this timestamp. If not present, will show the most recent transfers.
    :type timestamp: Optional[str]
    :param limit_transfers: The maximum number of transfers to return. Default is 10, max is 50.
    :type limit_transfers: Optional[int]
    :param show_completed_deposit_advances: Whether to display completed deposit advances. False by default. Must be set True to activate.
    :type show_completed_deposit_advances: Optional[int]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * type - Transfer type. Deposit or Withdrawal.
                      * status - Transfer status. Advanced or Complete.
                      * timestampms - The time that the trade was executed in milliseconds
                      * eid - Transfer event id
                      * advanceEid - Deposit advance event id
                      * currency - Currency code
                      * amount - The transfer amount
                      * method - Optional. When currency is a fiat currency, the method field will attempt to supply ACH, Wire, or SEN. If the transfer is an internal transfer between subaccounts the method field will return Internal.
                      * txHash - Optional. When currency is a cryptocurrency, supplies the transaction hash when available.
                      * outputIdx - Optional. When currency is a cryptocurrency, supplies the output index in the transaction when available.
                      * destination - Optional. When currency is a cryptocurrency, supplies the destination address when available.
                      * purpose - Optional. Administrative field used to supply a reason for certain types of advances.
    """
    url = URLS.transfers()
    payload = {
        "request": URLS.get_endpoint(url),
        "show_completed_deposit_advances": show_completed_deposit_advances
    }
    if timestamp:
        payload["timestamp"] = timestamp
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def get_deposit_addresses(network, timestamp=None,  jsonify=None):
    """ Gets a list of all deposit addresses.

    :param network: network can be bitcoin, ethereum, bitcoincash, litecoin, zcash, filecoin.
    :type network: str
    :param timestamp: Only returns addresses created on or after this timestamp.
    :type timestamp: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * address - String representation of the new cryptocurrency address.
                      * timestamp - Creation date of the address.
                      * label - Optional. if you provided a label when creating the address, it will be echoed back here.

    """
    url = URLS.deposit_addresses(network)
    payload = {
        "request": URLS.get_endpoint(url)
    }
    if timestamp:
        payload["timestamp"] = timestamp
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def get_approved_addresses(network, jsonify=None):
    """ Allows viewing of Approved Address list.

    :param network: The network of the approved address. Network can be bitcoin, ethereum, bitcoincash, litecoin, zcash, or filecoin
    :type network: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * approvedAddresses - Array of approved addresses on both the account and group level.
                        -- network - The network of the approved address. Network can be bitcoin, ethereum, bitcoincash, litecoin, zcash, or filecoin
                        -- scope - Will return the scope of the address as either "account" or "group"
                        -- label - The label assigned to the address
                        -- status - The status of the address that will return as "active", "pending-time" or "pending-mua". The remaining time is exactly 7 days after the initial request. "pending-mua" is for multi-user accounts and will require another administator or fund manager on the account to approve the address.
                        -- createdAt - UTC timestamp in millisecond of when the address was created.
                        -- address - The address on the approved address list.

    """
    url = URLS.approved_addresses(network)
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def withdraw_crypto_funds(currency_code, address, amount, jsonify=None):
    """ Before you can withdraw cryptocurrency funds to an approved address, you need three things:

        1. You must have an approved address list for your account
        2. The address you want to withdraw funds to needs to already be on that approved address list
        3. An API key with the Fund Manager role added

    :param currency_code: the three-letter currency code of a supported crypto-currency, e.g. btc or eth. 
    :type currency_code: str
    :param address: Standard string format of cryptocurrency address.
    :type address: str
    :param amount: 	Quoted decimal amount to withdraw.
    :type amount: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * address - Standard string format of the withdrawal destination address.
                      * amount - The withdrawal amount.
                      * txHash - Standard string format of the transaction hash of the withdrawal transaction. Only shown for ETH and GUSD withdrawals.
                      * withdrawalID - A unique ID for the withdrawal. Only shown for BTC, ZEC, LTC and BCH withdrawals.
                      * message - A human-readable English string describing the withdrawal. Only shown for BTC, ZEC, LTC and BCH withdrawals.

    """
    url = URLS.withdrawl_crypto(currency_code)
    payload = {
        "request": URLS.get_endpoint(url),
        "address": address,
        "amount": amount
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err



"""Contains functions for getting information related to the user account."""
import os
from uuid import uuid4

from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.profiles import *
from robin_stocks.robinhood.stocks import *
from robin_stocks.robinhood.urls import *


@login_required
def load_phoenix_account(info=None):
    """Returns unified information about your account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries of key/value pairs. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * account_buying_power
                      * cash_available_from_instant_deposits
                      * cash_held_for_currency_orders
                      * cash_held_for_dividends
                      * cash_held_for_equity_orders
                      * cash_held_for_options_collateral
                      * cash_held_for_orders
                      * crypto
                      * crypto_buying_power
                      * equities
                      * extended_hours_portfolio_equity
                      * instant_allocated
                      * levered_amount
                      * near_margin_call
                      * options_buying_power
                      * portfolio_equity
                      * portfolio_previous_close
                      * previous_close
                      * regular_hours_portfolio_equity
                      * total_equity
                      * total_extended_hours_equity
                      * total_extended_hours_market_value
                      * total_market_value
                      * total_regular_hours_equity
                      * total_regular_hours_market_value
                      * uninvested_cash
                      * withdrawable_cash

    """
    url = phoenix_url()
    data = request_get(url, 'regular')
    return(filter_data(data, info))

@login_required
def get_historical_portfolio(interval=None, span='week', bounds='regular',info=None):
    interval_check = ['5minute', '10minute', 'hour', 'day', 'week']
    span_check = ['day', 'week', 'month', '3month', 'year', '5year', 'all']
    bounds_check = ['extended', 'regular', 'trading']

    if interval not in interval_check:
        if interval is None and (bounds != 'regular' and span != 'all'):
            print ('ERROR: Interval must be None for "all" span "regular" bounds', file=get_output())
            return ([None])
        print(
            'ERROR: Interval must be "5minute","10minute","hour","day",or "week"', file=get_output())
        return([None])
    if span not in span_check:
        print('ERROR: Span must be "day","week","month","3month","year",or "5year"', file=get_output())
        return([None])
    if bounds not in bounds_check:
        print('ERROR: Bounds must be "extended","regular",or "trading"')
        return([None])
    if (bounds == 'extended' or bounds == 'trading') and span != 'day':
        print('ERROR: extended and trading bounds can only be used with a span of "day"', file=get_output())
        return([None])

    account = load_account_profile('account_number')
    url = portfolis_historicals_url(account)
    payload = {
        'interval': interval,
        'span': span,
        'bounds': bounds
    }
    data = request_get(url, 'regular', payload)

    return(filter_data(data, info))

@login_required
def get_all_positions(info=None):
    """Returns a list containing every position ever traded.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries of key/value pairs for each ticker. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * url
                      * instrument
                      * account
                      * account_number
                      * average_buy_price
                      * pending_average_buy_price
                      * quantity
                      * intraday_average_buy_price
                      * intraday_quantity
                      * shares_held_for_buys
                      * shares_held_for_sells
                      * shares_held_for_stock_grants
                      * shares_held_for_options_collateral
                      * shares_held_for_options_events
                      * shares_pending_from_options_events
                      * updated_at
                      * created_at

    """
    url = positions_url()
    data = request_get(url, 'pagination')

    return(filter_data(data, info))


@login_required
def get_open_stock_positions(account_number=None, info=None):
    """Returns a list of stocks that are currently held.

    :param acccount_number: the robinhood account number.
    :type acccount_number: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries of key/value pairs for each ticker. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * url
                      * instrument
                      * account
                      * account_number
                      * average_buy_price
                      * pending_average_buy_price
                      * quantity
                      * intraday_average_buy_price
                      * intraday_quantity
                      * shares_held_for_buys
                      * shares_held_for_sells
                      * shares_held_for_stock_grants
                      * shares_held_for_options_collateral
                      * shares_held_for_options_events
                      * shares_pending_from_options_events
                      * updated_at
                      * created_at

    """
    url = positions_url(account_number=account_number)
    payload = {'nonzero': 'true'}
    data = request_get(url, 'pagination', payload)

    return(filter_data(data, info))


@login_required
def get_dividends(info=None):
    """Returns a list of dividend trasactions that include information such as the percentage rate,
    amount, shares of held stock, and date paid.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries of key/value pairs for each divident payment. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * id
                      * url
                      * account
                      * instrument
                      * amount
                      * rate
                      * position
                      * withholding
                      * record_date
                      * payable_date
                      * paid_at
                      * state
                      * nra_withholding
                      * drip_enabled

    """
    url = dividends_url()
    data = request_get(url, 'pagination')

    return(filter_data(data, info))


@login_required
def get_total_dividends():
    """Returns a float number representing the total amount of dividends paid to the account.

    :returns: Total dollar amount of dividends paid to the account as a 2 precision float.

    """
    url = dividends_url()
    data = request_get(url, 'pagination')

    dividend_total = 0
    for item in data:
        dividend_total += float(item['amount']) if (item['state'] == 'paid' or item['state'] == 'reinvested') else 0
    return(dividend_total)


@login_required
def get_dividends_by_instrument(instrument, dividend_data):
    """Returns a dictionary with three fields when given the instrument value for a stock

    :param instrument: The instrument to get the dividend data.
    :type instrument: str
    :param dividend_data: The information returned by get_dividends().
    :type dividend_data: list
    :returns: dividend_rate       -- the rate paid for a single share of a specified stock \
              total_dividend      -- the total dividend paid based on total shares for a specified stock \
              amount_paid_to_date -- total amount earned by account for this particular stock
    """
    #global dividend_data
    try:
        data = list(
            filter(lambda x: x['instrument'] == instrument, dividend_data))

        dividend = float(data[0]['rate'])
        total_dividends = float(data[0]['amount'])
        total_amount_paid = float(sum([float(d['amount']) for d in data]))

        return {
            'dividend_rate': "{0:.2f}".format(dividend),
            'total_dividend': "{0:.2f}".format(total_dividends),
            'amount_paid_to_date': "{0:.2f}".format(total_amount_paid)
        }
    except:
        pass


@login_required
def get_notifications(info=None):
    """Returns a list of notifications.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each notification. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = notifications_url()
    data = request_get(url, 'pagination')

    return(filter_data(data, info))


@login_required
def get_latest_notification():
    """Returns the time of the latest notification.

    :returns: Returns a dictionary of key/value pairs. But there is only one key, 'last_viewed_at'

    """
    url = notifications_url(True)
    data = request_get(url)
    return(data)


@login_required
def get_wire_transfers(info=None):
    """Returns a list of wire transfers.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each wire transfer. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = wiretransfers_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_margin_calls(symbol=None):
    """Returns either all margin calls or margin calls for a specific stock.

    :param symbol: Will determine which stock to get margin calls for.
    :type symbol: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each margin call.

    """
    url = margin_url()
    if symbol:
        try:
            symbol = symbol.upper().strip()
        except AttributeError as message:
            print(message, file=get_output())
            return None
        payload = {'equity_instrument_id', id_for_stock(symbol)}
        data = request_get(url, 'results', payload)
    else:
        data = request_get(url, 'results')

    return(data)


@login_required
def withdrawl_funds_to_bank_account(ach_relationship, amount, info=None):
    """Submits a post request to withdraw a certain amount of money to a bank account.

    :param ach_relationship: The url of the bank account you want to withdrawl the money to.
    :type ach_relationship: str
    :param amount: The amount of money you wish to withdrawl.
    :type amount: float
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for the transaction.

    """
    url = banktransfers_url()
    payload = {
        "amount": amount,
        "direction": "withdraw",
        "ach_relationship": ach_relationship,
        "ref_id": str(uuid4())
    }
    data = request_post(url, payload)
    return(filter_data(data, info))


@login_required
def deposit_funds_to_robinhood_account(ach_relationship, amount, info=None):
    """Submits a post request to deposit a certain amount of money from a bank account to Robinhood.

    :param ach_relationship: The url of the bank account you want to deposit the money from.
    :type ach_relationship: str
    :param amount: The amount of money you wish to deposit.
    :type amount: float
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for the transaction.

    """
    url = banktransfers_url()
    payload = {
        "amount": amount,
        "direction": "deposit",
        "ach_relationship": ach_relationship,
        "ref_id": str(uuid4())
    }
    data = request_post(url, payload)
    return(filter_data(data, info))

@login_required
def get_linked_bank_accounts(info=None):
    """Returns all linked bank accounts.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each bank.

    """
    url = linked_url()
    data = request_get(url, 'results')
    return(filter_data(data, info))


@login_required
def get_bank_account_info(id, info=None):
    """Returns a single dictionary of bank information

    :param id: The bank id.
    :type id: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictinoary of key/value pairs for the bank. If info parameter is provided, \
    the value of the key that matches info is extracted.

    """
    url = linked_url(id)
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def unlink_bank_account(id):
    """Unlinks a bank account.

    :param id: The bank id.
    :type id: str
    :returns: Information returned from post request.

    """
    url = linked_url(id, True)
    data = request_post(url)
    return(data)


@login_required
def get_bank_transfers(direction=None, info=None):
    """Returns all bank transfers made for the account.

    :param direction: Possible values are 'received'. If left blank, function will return all withdrawls and deposits \
        that are initiated from Robinhood. If the value is 'received', funciton will return transfers intiated from \
        your bank rather than Robinhood.
    :type direction: Optional[str]
    :param info: Will filter the results to get a specific value. 'direction' gives if it was deposit or withdrawl.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each transfer. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = banktransfers_url(direction)
    data = request_get(url, 'pagination')
    return(filter_data(data, info))

@login_required
def get_card_transactions(cardType=None, info=None):
    """Returns all debit card transactions made on the account

    :param cardType: Will filter the card transaction types. Can be 'pending' or 'settled'.
    :type cardType: Optional[str]
    :param info: Will filter the results to get a specific value. 'direction' gives if it was debit or credit.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each transfer. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    payload = None
    if type:
        payload = { 'type': type }

    url = cardtransactions_url()
    data = request_get(url, 'pagination', payload)
    return(filter_data(data, info))

@login_required
def get_stock_loan_payments(info=None):
    """Returns a list of loan payments.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each payment. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = stockloan_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_margin_interest(info=None):
    """Returns a list of margin interest.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each interest. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = margininterest_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_subscription_fees(info=None):
    """Returns a list of subscription fees.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each fee. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = subscription_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_referrals(info=None):
    """Returns a list of referrals.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each referral. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = referral_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_day_trades(info=None):
    """Returns recent day trades.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each day trade. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    account = load_account_profile('account_number')
    url = daytrades_url(account)
    data = request_get(url, 'regular')
    return(filter_data(data, info))


@login_required
def get_documents(info=None):
    """Returns a list of documents that have been released by Robinhood to the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each document. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = documents_url()
    data = request_get(url, 'pagination')

    return(filter_data(data, info))


@login_required
def download_document(url, name=None, dirpath=None):
    """Downloads a document and saves as it as a PDF. If no name is given, document is saved as
    the name that Robinhood has for the document. If no directory is given, document is saved in the root directory of code.

    :param url: The url of the document. Can be found by using get_documents(info='download_url').
    :type url: str
    :param name: The name to save the document as.
    :type name: Optional[str]
    :param dirpath: The directory of where to save the document.
    :type dirpath: Optional[str]
    :returns: Returns the data from the get request.

    """
    data = request_document(url)

    print('Writing PDF...', file=get_output())
    if not name:
        name = url[36:].split('/', 1)[0]

    if dirpath:
        directory = dirpath
    else:
        directory = 'robin_documents/'

    filename = directory + name + ' .pdf'
    os.makedirs(os.path.dirname(filename), exist_ok=True)

    open(filename, 'wb').write(data.content)
    print('Done - Wrote file {}.pdf to {}'.format(name, os.path.abspath(filename)))

    return(data)


@login_required
def download_all_documents(doctype=None, dirpath=None):
    """Downloads all the documents associated with an account and saves them as a PDF.
    If no name is given, document is saved as a combination of the data of creation, type, and id.
    If no directory is given, document is saved in the root directory of code.

    :param doctype: The type of document to download, such as account_statement.
    :type doctype: Optional[str]
    :param dirpath: The directory of where to save the documents.
    :type dirpath: Optional[str]
    :returns: Returns the list of documents from get_documents(info=None)

    """
    documents = get_documents()

    downloaded_files = False
    if dirpath:
        directory = dirpath
    else:
        directory = 'robin_documents/'

    counter = 0
    for item in documents:
        if doctype == None:
            data = request_document(item['download_url'])
            if data:
                name = item['created_at'][0:10] + '-' + \
                    item['type'] + '-' + item['id']
                filename = directory + name + '.pdf'
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                open(filename, 'wb').write(data.content)
                downloaded_files = True
                counter += 1
                print('Writing PDF {}...'.format(counter), file=get_output())
        else:
            if item['type'] == doctype:
                data = request_document(item['download_url'])
                if data:
                    name = item['created_at'][0:10] + '-' + \
                        item['type'] + '-' + item['id']
                    filename = directory + name + '.pdf'
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                    open(filename, 'wb').write(data.content)
                    downloaded_files = True
                    counter += 1
                    print('Writing PDF {}...'.format(counter), file=get_output())

    if downloaded_files == False:
        print('WARNING: Could not find files of that doctype to download', file=get_output())
    else:
        if counter == 1:
            print('Done - wrote {} file to {}'.format(counter,
                                                      os.path.abspath(directory)), file=get_output())
        else:
            print('Done - wrote {} files to {}'.format(counter,
                                                       os.path.abspath(directory)), file=get_output())

    return(documents)


@login_required
def get_all_watchlists(info=None):
    """Returns a list of all watchlists that have been created. Everyone has a 'My First List' watchlist.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of the watchlists. Keywords are 'url', 'user', and 'name'.

    """
    url = watchlists_url()
    data = request_get(url, 'result')
    return(filter_data(data, info))


@login_required
def get_watchlist_by_name(name="My First List", info=None):
    """Returns a list of information related to the stocks in a single watchlist.

    :param name: The name of the watchlist to get data from.
    :type name: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries that contain the instrument urls and a url that references itself.

    """

    #Get id of requested watchlist
    all_watchlists = get_all_watchlists()
    watchlist_id = ''
    for wl in all_watchlists['results']:
        if wl['display_name'] == name:
            watchlist_id = wl['id']

    url = watchlists_url(name)
    data = request_get(url,'list_id',{'list_id':watchlist_id})
    return(filter_data(data, info))


@login_required
def post_symbols_to_watchlist(inputSymbols, name="My First List"):
    """Posts multiple stock tickers to a watchlist.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param name: The name of the watchlist to post data to.
    :type name: Optional[str]
    :returns: Returns result of the post request.

    """
    symbols = inputs_to_set(inputSymbols)
    ids = get_instruments_by_symbols(symbols, info='id')
    data = []
    #Get id of requested watchlist
    all_watchlists = get_all_watchlists()
    watchlist_id = ''
    for wl in all_watchlists['results']:
        if wl['display_name'] == name:
            watchlist_id = wl['id']

    for id in ids:
        payload = {
            watchlist_id: [{
                "object_type" : "instrument",
                "object_id" : id,
                "operation" : "create"
            }]
        }
        url = watchlists_url(name, True)
        data.append(request_post(url, payload, json=True))

    return(data)


@login_required
def delete_symbols_from_watchlist(inputSymbols, name="My First List"):
    """Deletes multiple stock tickers from a watchlist.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param name: The name of the watchlist to delete data from.
    :type name: Optional[str]
    :returns: Returns result of the delete request.

    """
    symbols = inputs_to_set(inputSymbols)
    ids = get_instruments_by_symbols(symbols, info='id')
    data = []

    #Get id of requested watchlist
    all_watchlists = get_all_watchlists()
    watchlist_id = ''
    for wl in all_watchlists['results']:
        if wl['display_name'] == name:
            watchlist_id = wl['id']

    for id in ids:
        payload = {
            watchlist_id: [{
                "object_type" : "instrument",
                "object_id" : id,
                "operation" : "delete"
            }]
        }
        url = watchlists_url(name, True)
        data.append(request_post(url, payload, json=True))

    return(data)


@login_required
def build_holdings(with_dividends=False):
    """Builds a dictionary of important information regarding the stocks and positions owned by the user.

    :param with_dividends: True if you want to include divident information.
    :type with_dividends: bool
    :returns: Returns a dictionary where the keys are the stock tickers and the value is another dictionary \
    that has the stock price, quantity held, equity, percent change, equity change, type, name, id, pe ratio, \
    percentage of portfolio, and average buy price.

    """
    positions_data = get_open_stock_positions()
    portfolios_data = load_portfolio_profile()
    accounts_data = load_account_profile()

    # user wants dividend information in their holdings
    if with_dividends is True:
        dividend_data = get_dividends()

    if not positions_data or not portfolios_data or not accounts_data:
        return({})

    if portfolios_data['extended_hours_equity'] is not None:
        total_equity = max(float(portfolios_data['equity']), float(
            portfolios_data['extended_hours_equity']))
    else:
        total_equity = float(portfolios_data['equity'])

    cash = "{0:.2f}".format(
        float(accounts_data['cash']) + float(accounts_data['uncleared_deposits']))

    holdings = {}
    for item in positions_data:
        # It is possible for positions_data to be [None]
        if not item:
            continue

        try:
            instrument_data = get_instrument_by_url(item['instrument'])
            symbol = instrument_data['symbol']
            fundamental_data = get_fundamentals(symbol)[0]

            price = get_latest_price(instrument_data['symbol'])[0]
            quantity = item['quantity']
            equity = float(item['quantity']) * float(price)
            equity_change = (float(quantity) * float(price)) - \
                (float(quantity) * float(item['average_buy_price']))
            percentage = float(item['quantity']) * float(price) * \
                100 / (float(total_equity) - float(cash))
            if (float(item['average_buy_price']) == 0.0):
                percent_change = 0.0
            else:
                percent_change = (float(
                    price) - float(item['average_buy_price'])) * 100 / float(item['average_buy_price'])
            if (float(item['intraday_average_buy_price']) == 0.0):
                intraday_percent_change = 0.0
            else:
                intraday_percent_change = (float(
                    price) - float(item['intraday_average_buy_price'])) * 100 / float(item['intraday_average_buy_price'])
            holdings[symbol] = ({'price': price})
            holdings[symbol].update({'quantity': quantity})
            holdings[symbol].update(
                {'average_buy_price': item['average_buy_price']})
            holdings[symbol].update({'equity': "{0:.2f}".format(equity)})
            holdings[symbol].update(
                {'percent_change': "{0:.2f}".format(percent_change)})
            holdings[symbol].update(
                {'intraday_percent_change': "{0:.2f}".format(intraday_percent_change)})
            holdings[symbol].update(
                {'equity_change': "{0:2f}".format(equity_change)})
            holdings[symbol].update({'type': instrument_data['type']})
            holdings[symbol].update(
                {'name': get_name_by_symbol(symbol)})
            holdings[symbol].update({'id': instrument_data['id']})
            holdings[symbol].update({'pe_ratio': fundamental_data['pe_ratio']})
            holdings[symbol].update(
                {'percentage': "{0:.2f}".format(percentage)})

            if with_dividends is True:
                # dividend_data was retrieved earlier
                holdings[symbol].update(get_dividends_by_instrument(
                    item['instrument'], dividend_data))

        except:
            pass

    return(holdings)


@login_required
def build_user_profile():
    """Builds a dictionary of important information regarding the user account.

    :returns: Returns a dictionary that has total equity, extended hours equity, cash, and divendend total.

    """
    user = {}

    portfolios_data = load_portfolio_profile()
    accounts_data = load_account_profile()

    if portfolios_data:
        user['equity'] = portfolios_data['equity']
        user['extended_hours_equity'] = portfolios_data['extended_hours_equity']

    if accounts_data:
        cash = "{0:.2f}".format(
            float(accounts_data['cash']) + float(accounts_data['uncleared_deposits']))
        user['cash'] = cash

    user['dividend_total'] = get_total_dividends()

    return(user)



""" Module contains all the API endpoints """
from enum import Enum, auto
from re import IGNORECASE, match, split

from robin_stocks.gemini.helper import get_sandbox_flag


class AutoName(Enum):
    """Automatically sets an enum value to be its name when using auto()"""

    def _generate_next_value_(name, start, count, last_values):
        return name


class Version(AutoName):
    """Enum for different version types"""
    v1 = auto()
    v2 = auto()


class URLS:
    """ Static class for holding all urls."""
    __base_url = "https://api.gemini.com"
    __base_sandbox_url = "https://api.sandbox.gemini.com"

    def __init__(self):
        raise NotImplementedError(
            "Cannot create instance of {0}".format(self.__class__.__name__))

    @classmethod
    def get_base_url(cls, version):
        if get_sandbox_flag():
            url = cls.__base_sandbox_url
        else:
            url = cls.__base_url

        return url + "/" + version.value + "/"

    @classmethod
    def get_endpoint(cls, url):
        if match(cls.__base_sandbox_url, url, IGNORECASE):
            _, end = split(cls.__base_sandbox_url, url, IGNORECASE)
        elif match(cls.__base_url, url, IGNORECASE):
            _, end = split(cls.__base_url, url, IGNORECASE)
        else:
            raise ValueError("The URL has the wrong base.")
        
        return end

    # account.py
    @classmethod
    def account_detail(cls):
        return cls.get_base_url(Version.v1) + "account"

    @classmethod
    def available_balances(cls):
        return cls.get_base_url(Version.v1) + "balances"

    @classmethod
    def notional_balances(cls):
        return cls.get_base_url(Version.v1) + "notionalbalances/usd"

    @classmethod
    def transfers(cls):
        return cls.get_base_url(Version.v1) + "transfers"

    @classmethod
    def deposit_addresses(cls, network):
        return cls.get_base_url(Version.v1) + "addresses/{0}".format(network)

    @classmethod
    def approved_addresses(cls, network):
        return cls.get_base_url(Version.v1) + "approvedAddresses/account/{0}".format(network)

    @classmethod
    def withdrawl_crypto(cls, currency_code):
        return cls.get_base_url(Version.v1) + "withdraw/{0}".format(currency_code)

    # authentication.py
    @classmethod
    def heartbeat(cls):
        return cls.get_base_url(Version.v1) + "heartbeat"

    # crypto.py
    @classmethod
    def pubticker(cls, ticker):
        return cls.get_base_url(Version.v1) + "pubticker/{0}".format(ticker)

    @classmethod
    def ticker(cls, ticker):
        return cls.get_base_url(Version.v2) + "ticker/{0}".format(ticker)

    @classmethod
    def symbols(cls):
        return cls.get_base_url(Version.v1) + "symbols"

    @classmethod
    def symbol_details(cls, ticker):
        return cls.get_base_url(Version.v1) + "symbols/details/{0}".format(ticker)

    @classmethod
    def notional_volume(cls):
        return cls.get_base_url(Version.v1) + "notionalvolume"

    @classmethod
    def trade_volume(cls):
        return cls.get_base_url(Version.v1) + "tradevolume"

    # orders.py
    @classmethod
    def mytrades(cls):
        return cls.get_base_url(Version.v1) + "mytrades"

    @classmethod
    def cancel_session_orders(cls):
        return cls.get_base_url(Version.v1) + "order/cancel/session"

    @classmethod
    def cancel_order(cls):
        return cls.get_base_url(Version.v1) + "order/cancel"

    @classmethod
    def order_status(cls):
        return cls.get_base_url(Version.v1) + "order/status"

    @classmethod
    def active_orders(cls):
        return cls.get_base_url(Version.v1) + "orders"

    @classmethod
    def cancel_active_orders(cls):
        return cls.get_base_url(Version.v1) + "order/cancel/all"

    @classmethod
    def order_new(cls):
        return cls.get_base_url(Version.v1) + "order/new"



"""Holds the session header and other global variables."""
import sys
import os

from requests import Session

# Keeps track on if the user is logged in or not.
LOGGED_IN = False
# The session object for making get and post requests.
SESSION = Session()
SESSION.headers = {
    "Accept": "*/*",
    "Accept-Encoding": "gzip,deflate,br",
    "Accept-Language": "en-US,en;q=1",
    "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
    "X-Robinhood-API-Version": "1.315.0",
    "Connection": "keep-alive",
    "User-Agent": "*"
}

#All print() statement direct their output to this stream
#by default, we use stdout which is the existing behavior
#but a client can change to any normal Python stream that
#print() accepts.  Common options are
#sys.stderr for standard error
#open(os.devnull,"w") for dev null
#io.StringIO() to go to a string for the client to inspect
OUTPUT=sys.stdout



"""Contains all the url endpoints for interacting with Robinhood API."""
from robin_stocks.robinhood.helper import id_for_chain, id_for_stock

# Login


def login_url():
    return('https://api.robinhood.com/oauth2/token/')


def challenge_url(challenge_id):
    return('https://api.robinhood.com/challenge/{0}/respond/'.format(challenge_id))

# Profiles


def account_profile_url(account_number=None):
    if account_number:
        return('https://api.robinhood.com/accounts/'+account_number)
    else:
        return('https://api.robinhood.com/accounts/')


def basic_profile_url():
    return('https://api.robinhood.com/user/basic_info/')


def investment_profile_url():
    return('https://api.robinhood.com/user/investment_profile/')


def portfolio_profile_url():
    return('https://api.robinhood.com/portfolios/')


def security_profile_url():
    return('https://api.robinhood.com/user/additional_info/')


def user_profile_url():
    return('https://api.robinhood.com/user/')

def portfolis_historicals_url(account_number):
    return('https://api.robinhood.com/portfolios/historicals/{0}/'.format(account_number))

# Stocks


def earnings_url():
    return('https://api.robinhood.com/marketdata/earnings/')


def events_url():
    return('https://api.robinhood.com/options/events/')


def fundamentals_url():
    return('https://api.robinhood.com/fundamentals/')


def historicals_url():
    return('https://api.robinhood.com/quotes/historicals/')


def instruments_url():
    return('https://api.robinhood.com/instruments/')


def news_url(symbol):
    return('https://api.robinhood.com/midlands/news/{0}/?'.format(symbol))


def popularity_url(symbol):
    return('https://api.robinhood.com/instruments/{0}/popularity/'.format(id_for_stock(symbol)))

def quotes_url():
    return('https://api.robinhood.com/quotes/')


def ratings_url(symbol):
    return('https://api.robinhood.com/midlands/ratings/{0}/'.format(id_for_stock(symbol)))


def splits_url(symbol):
    return('https://api.robinhood.com/instruments/{0}/splits/'.format(id_for_stock(symbol)))

# account

def phoenix_url():
    return('https://phoenix.robinhood.com/accounts/unified')

def positions_url(account_number=None):
    if account_number:
        return('https://api.robinhood.com/positions/?account_number='+account_number)
    else:
        return('https://api.robinhood.com/positions/')

def banktransfers_url(direction=None):
    if direction == 'received':
        return('https://api.robinhood.com/ach/received/transfers/')
    else:
        return('https://api.robinhood.com/ach/transfers/')

def cardtransactions_url():
   return('https://minerva.robinhood.com/history/transactions/')

def daytrades_url(account):
    return('https://api.robinhood.com/accounts/{0}/recent_day_trades/'.format(account))


def dividends_url():
    return('https://api.robinhood.com/dividends/')


def documents_url():
    return('https://api.robinhood.com/documents/')

def withdrawl_url(bank_id):
    return("https://api.robinhood.com/ach/relationships/{}/".format(bank_id))

def linked_url(id=None, unlink=False):
    if unlink:
        return('https://api.robinhood.com/ach/relationships/{0}/unlink/'.format(id))
    if id:
        return('https://api.robinhood.com/ach/relationships/{0}/'.format(id))
    else:
        return('https://api.robinhood.com/ach/relationships/')


def margin_url():
    return('https://api.robinhood.com/margin/calls/')


def margininterest_url():
    return('https://api.robinhood.com/cash_journal/margin_interest_charges/')


def notifications_url(tracker=False):
    if tracker:
        return('https://api.robinhood.com/midlands/notifications/notification_tracker/')
    else:
        return('https://api.robinhood.com/notifications/devices/')


def referral_url():
    return('https://api.robinhood.com/midlands/referral/')


def stockloan_url():
    return('https://api.robinhood.com/stock_loan/payments/')


def subscription_url():
    return('https://api.robinhood.com/subscription/subscription_fees/')


def wiretransfers_url():
    return('https://api.robinhood.com/wire/transfers')


def watchlists_url(name=None, add=False):
    if name:
        return('https://api.robinhood.com/midlands/lists/items/')
    else:
        return('https://api.robinhood.com/midlands/lists/default/')


# markets


def currency_url():
    return('https://nummus.robinhood.com/currency_pairs/')

def markets_url():
    return('https://api.robinhood.com/markets/')

def market_hours_url(market, date):
    return('https://api.robinhood.com/markets/{}/hours/{}/'.format(market, date))

def movers_sp500_url():
    return('https://api.robinhood.com/midlands/movers/sp500/')

def get_100_most_popular_url():
    return('https://api.robinhood.com/midlands/tags/tag/100-most-popular/')

def movers_top_url():
    return('https://api.robinhood.com/midlands/tags/tag/top-movers/')

def market_category_url(category):
    return('https://api.robinhood.com/midlands/tags/tag/{}/'.format(category))

# options


def aggregate_url():
    return('https://api.robinhood.com/options/aggregate_positions/')


def chains_url(symbol):
    return('https://api.robinhood.com/options/chains/{0}/'.format(id_for_chain(symbol)))


def option_historicals_url(id):
    return('https://api.robinhood.com/marketdata/options/historicals/{0}/'.format(id))


def option_instruments_url(id=None):
    if id:
        return('https://api.robinhood.com/options/instruments/{0}/'.format(id))
    else:
        return('https://api.robinhood.com/options/instruments/')


def option_orders_url(orderID=None, account_number=None):
    url = 'https://api.robinhood.com/options/orders/'
    if orderID:
        url += '{0}/'.format(orderID)
    if account_number:
        url += ('?account_numbers='+account_number)

    return url


def option_positions_url(account_number):
    if account_number:
        return('https://api.robinhood.com/options/positions/?account_numbers='+account_number)
    else:
        return('https://api.robinhood.com/options/positions/')


def marketdata_options_url():
    return('https://api.robinhood.com/marketdata/options/')

# pricebook


def marketdata_quotes_url(id):
    return ('https://api.robinhood.com/marketdata/quotes/{0}/'.format(id))


def marketdata_pricebook_url(id):
    return ('https://api.robinhood.com/marketdata/pricebook/snapshots/{0}/'.format(id))

# crypto


def order_crypto_url():
    return('https://nummus.robinhood.com/orders/')


def crypto_account_url():
    return('https://nummus.robinhood.com/accounts/')


def crypto_currency_pairs_url():
    return('https://nummus.robinhood.com/currency_pairs/')


def crypto_quote_url(id):
    return('https://api.robinhood.com/marketdata/forex/quotes/{0}/'.format(id))


def crypto_holdings_url():
    return('https://nummus.robinhood.com/holdings/')


def crypto_historical_url(id):
    return('https://api.robinhood.com/marketdata/forex/historicals/{0}/'.format(id))


def crypto_orders_url(orderID=None):
    if orderID:
        return('https://nummus.robinhood.com/orders/{0}/'.format(orderID))
    else:
        return('https://nummus.robinhood.com/orders/')


def crypto_cancel_url(id):
    return('https://nummus.robinhood.com/orders/{0}/cancel/'.format(id))

# orders


def cancel_url(url):
    return('https://api.robinhood.com/orders/{0}/cancel/'.format(url))


def option_cancel_url(id):
    return('https://api.robinhood.com/options/orders/{0}/cancel/'.format(id))


def orders_url(orderID=None):
    if orderID:
        return('https://api.robinhood.com/orders/{0}/'.format(orderID))
    else:
        return('https://api.robinhood.com/orders/')



from csv import writer
from datetime import date
from pathlib import Path
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.orders import *
from robin_stocks.robinhood.stocks import *
from robin_stocks.robinhood.crypto import *


def fix_file_extension(file_name):
    """ Takes a file extension and makes it end with .csv

    :param file_name: Name of the file.
    :type file_name: str
    :returns: Adds or replaces the file suffix with .csv and returns it as a string.

    """
    path = Path(file_name)
    path = path.with_suffix('.csv')
    return path.resolve()

def create_absolute_csv(dir_path, file_name, order_type):
    """ Creates a filepath given a directory and file name.

    :param dir_path: Absolute or relative path to the directory the file will be written.
    :type dir_path: str
    :param file_name: An optional argument for the name of the file. If not defined, filename will be stock_orders_{current date}
    :type file_name: str
    :param file_name: Will be 'stock', 'option', or 'crypto'
    :type file_name: str
    :returns: An absolute file path as a string.

    """
    path = Path(dir_path)
    directory = path.resolve()
    if not file_name:
        file_name = "{}_orders_{}.csv".format(order_type, date.today().strftime('%b-%d-%Y'))
    else:
        file_name = fix_file_extension(file_name)
    return(Path.joinpath(directory, file_name))


@login_required
def export_completed_stock_orders(dir_path, file_name=None):
    """Write all completed orders to a csv file

    :param dir_path: Absolute or relative path to the directory the file will be written.
    :type dir_path: str
    :param file_name: An optional argument for the name of the file. If not defined, filename will be stock_orders_{current date}
    :type file_name: Optional[str]

    """
    file_path = create_absolute_csv(dir_path, file_name, 'stock')
    all_orders = get_all_stock_orders()
    with open(file_path, 'w', newline='') as f:
        csv_writer = writer(f)
        csv_writer.writerow([
            'symbol',
            'date',
            'order_type',
            'side',
            'fees',
            'quantity',
            'average_price'
        ])
        for order in all_orders:
            # include candled order if partial executed
            if order['state'] == 'cancelled' and len(order['executions']) > 0:
                for partial in order['executions']:
                    csv_writer.writerow([
                        get_symbol_by_url(order['instrument']),
                        partial['timestamp'],
                        order['type'],
                        order['side'],
                        order['fees'],
                        partial['quantity'],
                        partial['price']
                    ])

            if order['state'] == 'filled' and order['cancel'] is None:
                csv_writer.writerow([
                    get_symbol_by_url(order['instrument']),
                    order['last_transaction_at'],
                    order['type'],
                    order['side'],
                    order['fees'],
                    order['quantity'],
                    order['average_price']
                ])
        f.close()
    
@login_required
def export_completed_crypto_orders(dir_path, file_name=None):
    """Write all completed crypto orders to a csv file

    :param dir_path: Absolute or relative path to the directory the file will be written.
    :type dir_path: str
    :param file_name: An optional argument for the name of the file. If not defined, filename will be crypto_orders_{current date}
    :type file_name: Optional[str]

    """
    file_path = create_absolute_csv(dir_path, file_name, 'crypto')
    all_orders = get_all_crypto_orders()
    
    with open(file_path, 'w', newline='') as f:
        csv_writer = writer(f)
        csv_writer.writerow([
            'symbol',
            'date',
            'order_type',
            'side',
            'fees',
            'quantity',
            'average_price'
        ])
        for order in all_orders:
            if order['state'] == 'filled' and order['cancel_url'] is None:
                try:
                    fees = order['fees']
                except KeyError:
                    fees = 0.0
                csv_writer.writerow([
                    get_crypto_quote_from_id(order['currency_pair_id'], 'symbol'),
                    order['last_transaction_at'],
                    order['type'],
                    order['side'],
                    fees,
                    order['quantity'],
                    order['average_price']
                ])
        f.close()


@login_required
def export_completed_option_orders(dir_path, file_name=None):
    """Write all completed option orders to a csv

        :param dir_path: Absolute or relative path to the directory the file will be written.
        :type dir_path: str
        :param file_name: An optional argument for the name of the file. If not defined, filename will be option_orders_{current date}
        :type file_name: Optional[str]

    """
    file_path = create_absolute_csv(dir_path, file_name, 'option')
    all_orders = get_all_option_orders()
    with open(file_path, 'w', newline='') as f:
        csv_writer = writer(f)
        csv_writer.writerow([
            'chain_symbol',
            'expiration_date',
            'strike_price',
            'option_type',
            'side',
            'order_created_at',
            'direction',
            'order_quantity',
            'order_type',
            'opening_strategy',
            'closing_strategy',
            'price',
            'processed_quantity'
        ])
        for order in all_orders:
            if order['state'] == 'filled':
                for leg in order['legs']:
                    instrument_data = request_get(leg['option'])
                    csv_writer.writerow([
                        order['chain_symbol'],
                        instrument_data['expiration_date'],
                        instrument_data['strike_price'],
                        instrument_data['type'],
                        leg['side'],
                        order['created_at'],
                        order['direction'],
                        order['quantity'],
                        order['type'],
                        order['opening_strategy'],
                        order['closing_strategy'],
                        order['price'],
                        order['processed_quantity']
                    ])
        f.close()



import os

import robin_stocks.tda as t
from dotenv import load_dotenv

load_dotenv()


class TestAuthentication:

    def test_login(self):
        t.login(os.environ['tda_encryption_passcode'])
        assert t.get_login_state()

class TestStocks:

    ticker = "TSLA"

    @classmethod
    def setup_class(cls):
        t.login(os.environ['tda_encryption_passcode'])

    def test_quote(self):
        resp, err = t.get_quote(self.ticker)
        data = resp.json()
        assert resp.status_code == 200
        assert err is None
        assert self.ticker in data




from .account import (build_holdings, build_user_profile,
                      delete_symbols_from_watchlist,
                      deposit_funds_to_robinhood_account,
                      download_all_documents, download_document,
                      get_all_positions, get_all_watchlists,
                      get_bank_account_info, get_bank_transfers,
                      get_card_transactions, get_day_trades, get_dividends,
                      get_dividends_by_instrument, get_documents,
                      get_historical_portfolio, get_latest_notification,
                      get_linked_bank_accounts, get_margin_calls,
                      get_margin_interest, get_notifications,
                      get_open_stock_positions, get_referrals,
                      get_stock_loan_payments, get_subscription_fees,
                      get_total_dividends, get_watchlist_by_name,
                      get_wire_transfers, load_phoenix_account,
                      post_symbols_to_watchlist, unlink_bank_account,
                      withdrawl_funds_to_bank_account)
from .authentication import login, logout
from .crypto import (get_crypto_currency_pairs, get_crypto_historicals,
                     get_crypto_info, get_crypto_positions, get_crypto_quote,
                     get_crypto_quote_from_id, load_crypto_profile)
from .export import (export_completed_crypto_orders,
                     export_completed_option_orders,
                     export_completed_stock_orders)
from .helper import (filter_data, get_output, request_delete, request_document,
                     request_get, request_post, set_output, update_session)
from .markets import (get_all_stocks_from_market_tag, get_currency_pairs,
                      get_market_hours, get_market_next_open_hours,
                      get_market_next_open_hours_after_date,
                      get_market_today_hours, get_markets, get_top_100,
                      get_top_movers, get_top_movers_sp500)
from .options import (find_options_by_expiration,
                      find_options_by_expiration_and_strike,
                      find_options_by_specific_profitability,
                      find_options_by_strike, find_tradable_options,
                      get_aggregate_positions, get_all_option_positions,
                      get_chains, get_market_options,
                      get_open_option_positions, get_option_historicals,
                      get_option_instrument_data,
                      get_option_instrument_data_by_id, get_option_market_data,
                      get_option_market_data_by_id)
from .orders import (cancel_all_crypto_orders, cancel_all_option_orders,
                     cancel_all_stock_orders, cancel_crypto_order,
                     cancel_option_order, cancel_stock_order,
                     find_stock_orders, get_all_crypto_orders,
                     get_all_open_crypto_orders, get_all_open_option_orders,
                     get_all_open_stock_orders, get_all_option_orders,
                     get_all_stock_orders, get_crypto_order_info,
                     get_option_order_info, get_stock_order_info, order,
                     order_buy_crypto_by_price, order_buy_crypto_by_quantity,
                     order_buy_crypto_limit, order_buy_crypto_limit_by_price,
                     order_buy_fractional_by_price,
                     order_buy_fractional_by_quantity, order_buy_limit,
                     order_buy_market, order_buy_option_limit,
                     order_buy_option_stop_limit, order_buy_stop_limit,
                     order_buy_stop_loss, order_buy_trailing_stop,
                     order_crypto, order_option_credit_spread,
                     order_option_debit_spread, order_option_spread,
                     order_sell_crypto_by_price, order_sell_crypto_by_quantity,
                     order_sell_crypto_limit, order_sell_crypto_limit_by_price,
                     order_sell_fractional_by_price,
                     order_sell_fractional_by_quantity, order_sell_limit,
                     order_sell_market, order_sell_option_limit,
                     order_sell_option_stop_limit, order_sell_stop_limit,
                     order_sell_stop_loss, order_sell_trailing_stop)
from .profiles import (load_account_profile, load_basic_profile,
                       load_investment_profile, load_portfolio_profile,
                       load_security_profile, load_user_profile)
from .stocks import (find_instrument_data, get_earnings, get_events,
                     get_fundamentals, get_instrument_by_url,
                     get_instruments_by_symbols, get_latest_price,
                     get_name_by_symbol, get_name_by_url, get_news,
                     get_pricebook_by_id, get_pricebook_by_symbol, get_quotes,
                     get_ratings, get_splits, get_stock_historicals,
                     get_stock_quote_by_id, get_stock_quote_by_symbol,
                     get_symbol_by_url)



"""Contains all functions for the purpose of logging in and out to Robinhood."""
import getpass
import os
import pickle
import random

from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *

def generate_device_token():
    """This function will generate a token used when loggin on.

    :returns: A string representing the token.

    """
    rands = []
    for i in range(0, 16):
        r = random.random()
        rand = 4294967296.0 * r
        rands.append((int(rand) >> ((3 & i) << 3)) & 255)

    hexa = []
    for i in range(0, 256):
        hexa.append(str(hex(i+256)).lstrip("0x").rstrip("L")[1:])

    id = ""
    for i in range(0, 16):
        id += hexa[rands[i]]

        if (i == 3) or (i == 5) or (i == 7) or (i == 9):
            id += "-"

    return(id)


def respond_to_challenge(challenge_id, sms_code):
    """This function will post to the challenge url.

    :param challenge_id: The challenge id.
    :type challenge_id: str
    :param sms_code: The sms code.
    :type sms_code: str
    :returns:  The response from requests.

    """
    url = challenge_url(challenge_id)
    payload = {
        'response': sms_code
    }
    return(request_post(url, payload))


def login(username=None, password=None, expiresIn=86400, scope='internal', by_sms=True, store_session=True, mfa_code=None, pickle_name=""):
    """This function will effectively log the user into robinhood by getting an
    authentication token and saving it to the session header. By default, it
    will store the authentication token in a pickle file and load that value
    on subsequent logins.

    :param username: The username for your robinhood account, usually your email.
        Not required if credentials are already cached and valid.
    :type username: Optional[str]
    :param password: The password for your robinhood account. Not required if
        credentials are already cached and valid.
    :type password: Optional[str]
    :param expiresIn: The time until your login session expires. This is in seconds.
    :type expiresIn: Optional[int]
    :param scope: Specifies the scope of the authentication.
    :type scope: Optional[str]
    :param by_sms: Specifies whether to send an email(False) or an sms(True)
    :type by_sms: Optional[boolean]
    :param store_session: Specifies whether to save the log in authorization
        for future log ins.
    :type store_session: Optional[boolean]
    :param mfa_code: MFA token if enabled.
    :type mfa_code: Optional[str]
    :param pickle_name: Allows users to name Pickle token file in order to switch
        between different accounts without having to re-login every time.
    :returns:  A dictionary with log in information. The 'access_token' keyword contains the access token, and the 'detail' keyword \
    contains information on whether the access token was generated or loaded from pickle file.

    """
    device_token = generate_device_token()
    home_dir = os.path.expanduser("~")
    data_dir = os.path.join(home_dir, ".tokens")
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    creds_file = "robinhood" + pickle_name + ".pickle"
    pickle_path = os.path.join(data_dir, creds_file)
    # Challenge type is used if not logging in with two-factor authentication.
    if by_sms:
        challenge_type = "sms"
    else:
        challenge_type = "email"

    url = login_url()
    payload = {
        'client_id': 'c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS',
        'expires_in': expiresIn,
        'grant_type': 'password',
        'password': password,
        'scope': scope,
        'username': username,
        'challenge_type': challenge_type,
        'device_token': device_token
    }

    if mfa_code:
        payload['mfa_code'] = mfa_code

    # If authentication has been stored in pickle file then load it. Stops login server from being pinged so much.
    if os.path.isfile(pickle_path):
        # If store_session has been set to false then delete the pickle file, otherwise try to load it.
        # Loading pickle file will fail if the acess_token has expired.
        if store_session:
            try:
                with open(pickle_path, 'rb') as f:
                    pickle_data = pickle.load(f)
                    access_token = pickle_data['access_token']
                    token_type = pickle_data['token_type']
                    refresh_token = pickle_data['refresh_token']
                    # Set device_token to be the original device token when first logged in.
                    pickle_device_token = pickle_data['device_token']
                    payload['device_token'] = pickle_device_token
                    # Set login status to True in order to try and get account info.
                    set_login_state(True)
                    update_session(
                        'Authorization', '{0} {1}'.format(token_type, access_token))
                    # Try to load account profile to check that authorization token is still valid.
                    res = request_get(
                        positions_url(), 'pagination', {'nonzero': 'true'}, jsonify_data=False)
                    # Raises exception is response code is not 200.
                    res.raise_for_status()
                    return({'access_token': access_token, 'token_type': token_type,
                            'expires_in': expiresIn, 'scope': scope, 'detail': 'logged in using authentication in {0}'.format(creds_file),
                            'backup_code': None, 'refresh_token': refresh_token})
            except:
                print(
                    "ERROR: There was an issue loading pickle file. Authentication may be expired - logging in normally.", file=get_output())
                set_login_state(False)
                update_session('Authorization', None)
        else:
            os.remove(pickle_path)

    # Try to log in normally.
    if not username:
        username = input("Robinhood username: ")
        payload['username'] = username

    if not password:
        password = getpass.getpass("Robinhood password: ")
        payload['password'] = password

    data = request_post(url, payload)
    # Handle case where mfa or challenge is required.
    if data:
        if 'mfa_required' in data:
            mfa_token = input("Please type in the MFA code: ")
            payload['mfa_code'] = mfa_token
            res = request_post(url, payload, jsonify_data=False)
            while (res.status_code != 200):
                mfa_token = input(
                    "That MFA code was not correct. Please type in another MFA code: ")
                payload['mfa_code'] = mfa_token
                res = request_post(url, payload, jsonify_data=False)
            data = res.json()
        elif 'challenge' in data:
            challenge_id = data['challenge']['id']
            sms_code = input('Enter Robinhood code for validation: ')
            res = respond_to_challenge(challenge_id, sms_code)
            while 'challenge' in res and res['challenge']['remaining_attempts'] > 0:
                sms_code = input('That code was not correct. {0} tries remaining. Please type in another code: '.format(
                    res['challenge']['remaining_attempts']))
                res = respond_to_challenge(challenge_id, sms_code)
            update_session(
                'X-ROBINHOOD-CHALLENGE-RESPONSE-ID', challenge_id)
            data = request_post(url, payload)
        # Update Session data with authorization or raise exception with the information present in data.
        if 'access_token' in data:
            token = '{0} {1}'.format(data['token_type'], data['access_token'])
            update_session('Authorization', token)
            set_login_state(True)
            data['detail'] = "logged in with brand new authentication code."
            if store_session:
                with open(pickle_path, 'wb') as f:
                    pickle.dump({'token_type': data['token_type'],
                                 'access_token': data['access_token'],
                                 'refresh_token': data['refresh_token'],
                                 'device_token': payload['device_token']}, f)
        else:
            raise Exception(data['detail'])
    else:
        raise Exception('Error: Trouble connecting to robinhood API. Check internet connection.')
    return(data)


@login_required
def logout():
    """Removes authorization from the session header.

    :returns: None

    """
    set_login_state(False)
    update_session('Authorization', None)




from base64 import b64encode
from datetime import datetime
from hashlib import sha384
from hmac import new
from json import dumps
from random import random
from time import mktime

from robin_stocks.gemini.helper import (format_inputs, get_secret_key, get_nonce,
                                        increment_nonce, login_required,
                                        request_post, set_secret_key,
                                        set_login_state, update_session)
from robin_stocks.gemini.urls import URLS


def login(api_key, secret_key):
    """ Set the authorization token so the API can be used.
    """
    update_session("X-GEMINI-APIKEY", api_key)
    set_secret_key(secret_key.encode())
    set_login_state(True)


def logout():
    """ Removes the API and Secret key from session and global variables.
    """
    update_session("X-GEMINI-APIKEY", "")
    set_secret_key("".encode())
    set_login_state(False)


def generate_signature(payload):
    """ Generate session header information needed to process Private API requests.

    :param payload: Dictionary of parameters to pass to encode.
    :type payload: dict

    """
    gemini_api_secret = get_secret_key()
    t = datetime.now()
    payload["nonce"] = str(int(mktime(t.timetuple())*1000) + get_nonce())
    encoded_payload = dumps(payload).encode()
    b64 = b64encode(encoded_payload)
    signature = new(gemini_api_secret, b64, sha384).hexdigest()
    update_session("X-GEMINI-PAYLOAD", b64)
    update_session("X-GEMINI-SIGNATURE", signature)
    increment_nonce()


def generate_order_id():
    """ This function will generate a token used when placing orders.

    :returns: A string representing the token.

    """
    rands = []
    for i in range(0, 16):
        r = random()
        rand = 4294967296.0 * r
        rands.append((int(rand) >> ((3 & i) << 3)) & 255)

    hexa = []
    for i in range(0, 256):
        hexa.append(str(hex(i+256)).lstrip("0x").rstrip("L")[1:])

    id = ""
    for i in range(0, 16):
        id += hexa[rands[i]]

        if (i == 3) or (i == 5) or (i == 7) or (i == 9):
            id += "-"

    return(id)


@login_required
@format_inputs
def heartbeat(jsonify=None):
    """ Generate a heartbeat response to keep a session alive.

    :returns: {"result": "ok"}

    """
    url = URLS.heartbeat()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err



from robin_stocks.gemini.authentication import generate_signature
from robin_stocks.gemini.helper import (format_inputs, login_required,
                                        request_get, request_post)
from robin_stocks.gemini.urls import URLS


@format_inputs
def get_pubticker(ticker, jsonify=None):
    """ Gets the pubticker information for a crypto.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * bid - The highest bid currently available
                      * ask - The lowest ask currently available
                      * last - The price of the last executed trade
                      * volume - Information about the 24 hour volume on the exchange

    """
    url = URLS.pubticker(ticker)
    data, error = request_get(url, None, jsonify)
    return data, error


@format_inputs
def get_ticker(ticker, jsonify=None):
    """ Gets the recent trading information for a crypto.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * symbol - BTCUSD etc.
                      * open - Open price from 24 hours ago
                      * high - High price from 24 hours ago
                      * low - Low price from 24 hours ago
                      * close - Close price (most recent trade)
                      * changes - Hourly prices descending for past 24 hours
                      * bid - Current best bid
                      * ask - Current best offer


    """
    url = URLS.ticker(ticker)
    data, error = request_get(url, None, jsonify)
    return data, error


@format_inputs
def get_symbols(jsonify=None):
    """ Gets a list of all available crypto tickers.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of strings and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.symbols()
    data, error = request_get(url, None, jsonify)
    return data, error


@format_inputs
def get_symbol_details(ticker, jsonify=None):
    """ Gets detailed information for a crypto.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * symbol - BTCUSD etc.
                      * base_currency - CCY1 or the top currency. (ie BTC in BTCUSD)
                      * quote_currency - CCY2 or the quote currency. (ie USD in BTCUSD)
                      * tick_size - The number of decimal places in the quote_currency
                      * quote_increment - The number of decimal places in the base_currency
                      * min_order_size - The minimum order size in base_currency units.
                      * status - Status of the current order book. Can be open, closed, cancel_only, post_only, limit_only.


    """
    url = URLS.symbol_details(ticker)
    data, error = request_get(url, None, jsonify)
    return data, error


@login_required
@format_inputs
def get_notional_volume(jsonify=None):
    """ Gets information about notional volume

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.               
    :Dictionary Keys: * date - UTC date in yyyy-MM-dd format
                      * last_updated_ms - Unix timestamp in millisecond of the last update
                      * web_maker_fee_bps - Integer value representing the maker fee for all symbols in basis point for web orders
                      * web_taker_fee_bps - Integer value representing the taker fee for all symbols in basis point for web orders
                      * web_auction_fee_bps - Integer value representing the auction fee for all symbols in basis point for web orders
                      * api_maker_fee_bps - Integer value representing the maker fee for all symbols in basis point for API orders
                      * api_taker_fee_bps - Integer value representing the taker fee for all symbols in basis point for API orders
                      * api_auction_fee_bps - Integer value representing the auction fee for all symbols in basis point for API orders
                      * fix_maker_fee_bps - Integer value representing the maker fee for all symbols in basis point for FIX orders
                      * fix_taker_fee_bps - Integer value representing the taker fee for all symbols in basis point for FIX orders
                      * fix_auction_fee_bps - Integer value representing the auction fee for all symbols in basis point for FIX orders
                      * block_maker_fee_bps - Integer value representing the maker fee for all symbols in basis point for block orders
                      * block_taker_fee_bps - Integer value representing the taker fee for all symbols in basis point for block orders
                      * notional_30d_volume - Maker plus taker trading volume for the past 30 days, including auction volume
                      * notional_1d_volume - A list of 1 day notional volume for the past 30 days
    """
    url = URLS.notional_volume()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def get_trade_volume(jsonify=None):
    """ Gets information about trade volume. The response will be an array of up to 30 days of trade volume for each symbol.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.               
    :Dictionary Keys: * symbol - The symbol.
                      * base_currency - quantity is denominated in this currency.
                      * notional_currency - price is denominated as the amount of notional currency per one unit of base currency. Notional values are denominated in this currency.
                      * data_date - UTC date in yyyy-MM-dd format.
                      * total_volume_base - Total trade volume for this day.
                      * maker_buy_sell_ratio - Maker buy/sell ratio is the proportion of maker base volume on trades where the account was on the buy side versus all maker trades. If there is no maker base volume on the buy side, then this value is 0.
                      * buy_maker_base - Quantity for this day where the account was a maker on the buy side of the trade.
                      * buy_maker_notional - Notional value for this day where the account was a maker on the buy side of the trade.
                      * buy_maker_count - Number of trades for this day where the account was a maker on the buy side of the trade.
                      * sell_maker_base - Quantity for this day where the account was a maker on the sell side of the trade.
                      * sell_maker_notional - Notional value for this day where the account was a maker on the sell side of the trade.
                      * sell_maker_count - Number of trades for this day where the account was a maker on the sell side of the trade.
                      * buy_taker_base- Quantity for this day where the account was a taker on the buy side of the trade.
                      * buy_taker_notional - Notional value for this day where the account was a taker on the buy side of the trade.
                      * buy_taker_count - Number of trades for this day where the account was a taker on the buy side of the trade.
                      * sell_taker_base - Quantity for this day where the account was a taker on the sell side of the trade.
                      * sell_taker_notional - Notional value for this day where the account was a taker on the sell side of the trade.
                      * sell_taker_count - Number of trades for this day where the account was a taker on the sell side of the trade.
                      
    """
    url = URLS.trade_volume()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


def get_price(ticker, side):
    """ Returns either the bid or the ask price as a string.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param side: Either 'buy' or 'sell'.
    :type side: str
    :returns: Returns the bid or ask price as a string.

    """
    data, _ = get_pubticker(ticker, jsonify=True)
    if side == "buy":
        return data["ask"]
    else:
        return data["bid"]



"""Contains functions to get information about crypto-currencies."""
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *

@login_required
def load_crypto_profile(info=None):
    """Gets the information associated with the crypto account.

    :param info: The name of the key whose value is to be returned from the function.
    :type info: Optional[str]
    :returns: [dict] The function returns a dictionary of key/value pairs. \
    If a string is passed in to the info parameter, then the function will return \
    a string corresponding to the value of the key whose name matches the info parameter.
    :Dictionary Keys: * apex_account_number
                      * created_at
                      * id
                      * rhs_account_number
                      * status
                      * status_reason_code
                      * updated_at
                      * user_id

    """
    url = crypto_account_url()
    data = request_get(url, 'indexzero')
    return(filter_data(data, info))


@login_required
def get_crypto_positions(info=None):
    """Returns crypto positions for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries of key/value pairs for each option. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.
    :Dictionary Keys: * account_id
                      * cost_basis
                      * created_at
                      * currency
                      * id
                      * quantity
                      * quantity_available
                      * quantity_held_for_buy
                      * quantity_held_for_sell
                      * updated_at

    """
    url = crypto_holdings_url()
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


def get_crypto_currency_pairs(info=None):
    """Gets a list of all the cypto currencies that you can trade.

    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * asset_currency
                      * display_only
                      * id
                      * max_order_size
                      * min_order_size
                      * min_order_price_increment
                      * min_order_quantity_increment
                      * name
                      * quote_currency
                      * symbol
                      * tradability

    """
    url = crypto_currency_pairs_url()
    data = request_get(url, 'results')
    return(filter_data(data, info))


def get_crypto_info(symbol, info=None):
    """Gets information about a crpyto currency.

    :param symbol: The crypto ticker.
    :type symbol: str
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [dict] If info parameter is left as None then will return a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a strings representing the value of the key.
    :Dictionary Keys: * asset_currency
                      * display_only
                      * id
                      * max_order_size
                      * min_order_size
                      * min_order_price_increment
                      * min_order_quantity_increment
                      * name
                      * quote_currency
                      * symbol
                      * tradability

    """
    url = crypto_currency_pairs_url()
    data = request_get(url, 'results')
    data = [x for x in data if x['asset_currency']['code'] == symbol]
    if len(data) > 0:
        data = data[0]
    else:
        data = None
    return(filter_data(data, info))


SYMBOL_TO_ID_CACHE = {}
def get_crypto_id(symbol):
    """Gets the Robinhood ID of the given cryptocurrency used to make trades.
    This function uses an in-memory cache of the IDs to save a network round-trip when possible.

    :param symbol: The crypto ticker.
    :type symbol: str
    :returns: [str] The symbol's Robinhood ID.
    """
    if symbol in SYMBOL_TO_ID_CACHE:
        return SYMBOL_TO_ID_CACHE[symbol]

    id = get_crypto_info(symbol, 'id')
    if id:
        SYMBOL_TO_ID_CACHE[symbol] = id
    return id


@login_required
def get_crypto_quote(symbol, info=None):
    """Gets information about a crypto including low price, high price, and open price

    :param symbol: The crypto ticker.
    :type symbol: str
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [dict] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * ask_price
                      * bid_price
                      * high_price
                      * id
                      * low_price
                      * mark_price
                      * open_price
                      * symbol
                      * volume
 
    """
    id = get_crypto_info(symbol, info='id')
    url = crypto_quote_url(id)
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def get_crypto_quote_from_id(id, info=None):
    """Gets information about a crypto including low price, high price, and open price. Uses the id instead of crypto ticker.

    :param id: The id of a crypto.
    :type id: str
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [dict] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * ask_price
                      * bid_price
                      * high_price
                      * id
                      * low_price
                      * mark_price
                      * open_price
                      * symbol
                      * volume

    """
    url = crypto_quote_url(id)
    data = request_get(url)
    return(filter_data(data, info))


@login_required
def get_crypto_historicals(symbol, interval='hour', span='week', bounds='24_7', info=None):
    """Gets historical information about a crypto including open price, close price, high price, and low price.

    :param symbol: The crypto ticker.
    :type symbol: str
    :param interval: The time between data points. Can be '15second', '5minute', '10minute', 'hour', 'day', or 'week'. Default is 'hour'.
    :type interval: str
    :param span: The entire time frame to collect data points. Can be 'hour', 'day', 'week', 'month', '3month', 'year', or '5year'. Default is 'week'
    :type span: str
    :param bound: The times of day to collect data points. 'Regular' is 6 hours a day, 'trading' is 9 hours a day, \
    'extended' is 16 hours a day, '24_7' is 24 hours a day. Default is '24_7'
    :type bound: str
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * begins_at
                      * open_price
                      * close_price
                      * high_price
                      * low_price
                      * volume
                      * session
                      * interpolated
                      * symbol

    """
    interval_check = ['15second', '5minute', '10minute', 'hour', 'day', 'week']
    span_check = ['hour', 'day', 'week', 'month', '3month', 'year', '5year']
    bounds_check = ['24_7', 'extended', 'regular', 'trading']

    if interval not in interval_check:
        print(
            'ERROR: Interval must be "15second","5minute","10minute","hour","day",or "week"', file=get_output())
        return([None])
    if span not in span_check:
        print('ERROR: Span must be "hour","day","week","month","3month","year",or "5year"', file=get_output())
        return([None])
    if bounds not in bounds_check:
        print('ERROR: Bounds must be "24_7","extended","regular",or "trading"', file=get_output())
        return([None])
    if (bounds == 'extended' or bounds == 'trading') and span != 'day':
        print('ERROR: extended and trading bounds can only be used with a span of "day"', file=get_output())
        return([None])


    symbol = inputs_to_set(symbol)
    id = get_crypto_info(symbol[0], info='id')
    url = crypto_historical_url(id)
    payload = {'interval': interval,
               'span': span,
               'bounds': bounds}
    data = request_get(url, 'regular', payload)

    histData = []
    cryptoSymbol = data['symbol']
    for subitem in data['data_points']:
        subitem['symbol'] = cryptoSymbol
        histData.append(subitem)

    return(filter_data(histData, info))



import os

import robin_stocks.gemini as g
from dotenv import load_dotenv

load_dotenv()


class TestAuthentication:

    def test_login(self):
        g.login(os.environ['gemini_account_key'],
                os.environ['gemini_account_secret'])
        assert g.get_login_state()

    def test_logout(self):
        g.login(os.environ['gemini_account_key'],
                os.environ['gemini_account_secret'])
        g.logout()
        assert not g.get_login_state()

    def test_heartbeat(self):
        g.login(os.environ['gemini_account_key'],
                os.environ['gemini_account_secret'])
        response, err = g.heartbeat()
        data = response.json()
        assert err == None
        assert response.status_code == 200
        assert data["result"] == "ok"


class TestCrypto:

    ticker = "btcusd"

    def test_pubticker_btc(self):
        response, err = g.get_pubticker(self.ticker)
        data = response.json()
        assert err == None
        assert response.status_code == 200
        assert "bid" in data
        assert "ask" in data
        assert "volume" in data
        assert "last" in data

    def test_get_symbols(self):
        response, err = g.get_symbols()
        data = response.json()
        assert err == None
        assert response.status_code == 200
        assert len(data) > 1
        assert self.ticker in data


class TestOrders:

    ticker = "ethusd"

    @classmethod
    def setup_class(cls):
        g.use_sand_box_urls(True)
        g.login(os.environ['gemini_sandbox_key'],
                os.environ['gemini_sandbox_secret'])

    @classmethod
    def teardown_class(cls):
        g.use_sand_box_urls(False)

    def test_mytrades(self):
        response, err = g.get_trades_for_crypto("btcusd")
        assert err == None
        assert response.status_code == 200


class TestAccount:

    @classmethod
    def setup_class(cls):
        g.login(os.environ['gemini_account_key'],
                os.environ['gemini_account_secret'])

    def test_account_detail(self):
        response, err = g.get_account_detail()
        assert err == None
        assert response.status_code == 200



"""Contains information in regards to stocks."""
from functools import cache

from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *

def get_quotes(inputSymbols, info=None):
    """Takes any number of stock tickers and returns information pertaining to its price.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument

    """
    symbols = inputs_to_set(inputSymbols)
    url = quotes_url()
    payload = {'symbols': ','.join(symbols)}
    data = request_get(url, 'results', payload)

    if (data == None or data == [None]):
        return data

    for count, item in enumerate(data):
        if item is None:
            print(error_ticker_does_not_exist(symbols[count]), file=get_output())

    data = [item for item in data if item is not None]

    return(filter_data(data, info))


def get_fundamentals(inputSymbols, info=None):
    """Takes any number of stock tickers and returns fundamental information
    about the stock such as what sector it is in, a description of the company, dividend yield, and market cap.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * open
                      * high
                      * low
                      * volume
                      * average_volume_2_weeks
                      * average_volume
                      * high_52_weeks
                      * dividend_yield
                      * float
                      * low_52_weeks
                      * market_cap
                      * pb_ratio
                      * pe_ratio
                      * shares_outstanding
                      * description
                      * instrument
                      * ceo
                      * headquarters_city
                      * headquarters_state
                      * sector
                      * industry
                      * num_employees
                      * year_founded
                      * symbol

    """ 
    symbols = inputs_to_set(inputSymbols)
    url = fundamentals_url()
    payload = {'symbols': ','.join(symbols)}
    data = request_get(url, 'results', payload)

    if (data == None or data == [None]):
        return data

    for count, item in enumerate(data):
        if item is None:
            print(error_ticker_does_not_exist(symbols[count]), file=get_output())
        else:
            item['symbol'] = symbols[count]

    data = [item for item in data if item is not None]

    return(filter_data(data, info))


def get_instruments_by_symbols(inputSymbols, info=None):
    """Takes any number of stock tickers and returns information held by the market
    such as ticker name, bloomberg id, and listing date.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] If info parameter is left as None then the list will a dictionary of key/value pairs for each ticker. \
    Otherwise, it will be a list of strings where the strings are the values of the key that corresponds to info.
    :Dictionary Keys: * id
                      * url
                      * quote
                      * fundamentals
                      * splits
                      * state
                      * market
                      * simple_name
                      * name
                      * tradeable
                      * tradability
                      * symbol
                      * bloomberg_unique
                      * margin_initial_ratio
                      * maintenance_ratio
                      * country
                      * day_trade_ratio
                      * list_date
                      * min_tick_size
                      * type
                      * tradable_chain_id
                      * rhs_tradability
                      * fractional_tradability
                      * default_collar_fraction

    """ 
    symbols = inputs_to_set(inputSymbols)
    url = instruments_url()
    data = []
    for item in symbols:
        payload = {'symbol': item}
        itemData = request_get(url, 'indexzero', payload)

        if itemData:
            data.append(itemData)
        else:
            print(error_ticker_does_not_exist(item), file=get_output())

    return(filter_data(data, info))


def get_instrument_by_url(url, info=None):
    """Takes a single url for the stock. Should be located at ``https://api.robinhood.com/instruments/<id>`` where <id> is the
    id of the stock.

    :param url: The url of the stock. Can be found in several locations including \
    in the dictionary returned from get_instruments_by_symbols(inputSymbols,info=None)
    :type url: str
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [dict or str] If info parameter is left as None then will return a dictionary of key/value pairs for a specific url. \
    Otherwise, it will be the string value of the key that corresponds to info.
    :Dictionary Keys: * id
                      * url
                      * quote
                      * fundamentals
                      * splits
                      * state
                      * market
                      * simple_name
                      * name
                      * tradeable
                      * tradability
                      * symbol
                      * bloomberg_unique
                      * margin_initial_ratio
                      * maintenance_ratio
                      * country
                      * day_trade_ratio
                      * list_date
                      * min_tick_size
                      * type
                      * tradable_chain_id
                      * rhs_tradability
                      * fractional_tradability
                      * default_collar_fraction

    """
    data = request_get(url, 'regular')

    return(filter_data(data, info))


def get_latest_price(inputSymbols, priceType=None, includeExtendedHours=True):
    """Takes any number of stock tickers and returns the latest price of each one as a string.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param priceType: Can either be 'ask_price' or 'bid_price'. If this parameter is set, then includeExtendedHours is ignored.
    :type priceType: str
    :param includeExtendedHours: Leave as True if you want to get extendedhours price if available. \
    False if you only want regular hours price, even after hours.
    :type includeExtendedHours: bool
    :returns: [list] A list of prices as strings.

    """ 
    symbols = inputs_to_set(inputSymbols)
    quote = get_quotes(symbols)

    prices = []
    for item in quote:
        if item:
            if priceType == 'ask_price':
                prices.append(item['ask_price'])
            elif priceType == 'bid_price':
                prices.append(item['bid_price'])
            else:
                if priceType:
                    print('WARNING: priceType should be "ask_price" or "bid_price". You entered "{0}"'.format(priceType), file=get_output())
                if item['last_extended_hours_trade_price'] is None or not includeExtendedHours:
                    prices.append(item['last_trade_price'])
                else:
                    prices.append(item['last_extended_hours_trade_price'])
        else:
            prices.append(None)
    return(prices)

@cache
@convert_none_to_string
def get_name_by_symbol(symbol):
    """Returns the name of a stock from the stock ticker.

    :param symbol: The ticker of the stock as a string.
    :type symbol: str
    :returns: [str] Returns the simple name of the stock. If the simple name does not exist then returns the full name.

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = instruments_url()
    payload = {'symbol': symbol}
    data = request_get(url, 'indexzero', payload)
    if not data:
        return(None)
    # If stock doesn't have a simple name attribute then get the full name.
    filter = filter_data(data, info='simple_name')
    if not filter or filter == "":
        filter = filter_data(data, info='name')
    return(filter)


@cache
@convert_none_to_string
def get_name_by_url(url):
    """Returns the name of a stock from the instrument url. Should be located at ``https://api.robinhood.com/instruments/<id>``
    where <id> is the id of the stock.

    :param url: The url of the stock as a string.
    :type url: str
    :returns: [str] Returns the simple name of the stock. If the simple name does not exist then returns the full name.

    """
    data = request_get(url)
    if not data:
        return(None)
    # If stock doesn't have a simple name attribute then get the full name.
    filter = filter_data(data, info='simple_name')
    if not filter or filter == "":
        filter = filter_data(data, info='name')
    return(filter)


@cache
@convert_none_to_string
def get_symbol_by_url(url):
    """Returns the symbol of a stock from the instrument url. Should be located at ``https://api.robinhood.com/instruments/<id>``
    where <id> is the id of the stock.

    :param url: The url of the stock as a string.
    :type url: str
    :returns: [str] Returns the ticker symbol of the stock.

    """
    data = request_get(url)
    return filter_data(data, info='symbol')

@convert_none_to_string
def get_ratings(symbol, info=None):
    """Returns the ratings for a stock, including the number of buy, hold, and sell ratings.

    :param symbol: The stock ticker.
    :type symbol: str
    :param info: Will filter the results to contain a dictionary of values that correspond to the key that matches info. \
    Possible values are summary, ratings, and instrument_id
    :type info: Optional[str]
    :returns: [dict] If info parameter is left as None then the list will contain a dictionary of key/value pairs for each ticker. \
    Otherwise, it will contain the values that correspond to the keyword that matches info.
    :Dictionary Keys: * summary - value is a dictionary
                      * ratings - value is a list of dictionaries
                      * instrument_id - value is a string
                      * ratings_published_at - value is a string

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = ratings_url(symbol)
    data = request_get(url)
    if not data:
        return(data)

    if (len(data['ratings']) == 0):
        return(data)
    else:
        for item in data['ratings']:
            oldText = item['text']
            item['text'] = oldText.encode('UTF-8')

    return(filter_data(data, info))
    

def get_events(symbol, info=None):
    """Returns the events related to a stock that the user owns. For example, if you owned options for USO and that stock \
    underwent a stock split resulting in you owning shares of newly created USO1, then that event will be returned when calling \
    get_events('uso1')

    :param symbol: The stock ticker.
    :type symbol: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] If the info parameter is provided, then the function will extract the value of the key \
    that matches the info parameter. Otherwise, the whole dictionary is returned.
    :Dictionary Keys: * account
                      * cash_component
                      * chain_id
                      * created_at
                      * direction
                      * equity_components
                      * event_date
                      * id
                      * option
                      * position
                      * quantity
                      * state
                      * total_cash_amount
                      * type
                      * underlying_price
                      * updated_at

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    payload = {'equity_instrument_id': id_for_stock(symbol)}
    url = events_url()
    data = request_get(url, 'results', payload)

    return(filter_data(data, info))


def get_earnings(symbol, info=None):
    """Returns the earnings for the different financial quarters.

    :param symbol: The stock ticker.
    :type symbol: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries. If info parameter is provided, \
    a list of strings is returned where the strings are the value \
    of the key that matches info.
    :Dictionary Keys: * symbol
                      * instrument
                      * year
                      * quarter
                      * eps
                      * report
                      * call

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = earnings_url()
    payload = {'symbol': symbol}
    data = request_get(url, 'results', payload)

    return(filter_data(data, info))


def get_news(symbol, info=None):
    """Returns news stories for a stock.

    :param symbol: The stock ticker.
    :type symbol: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries. If info parameter is provided, \
    a list of strings is returned where the strings are the value \
    of the key that matches info.
    :Dictionary Keys: * api_source
                      * author
                      * num_clicks
                      * preview_image_url
                      * published_at
                      * relay_url
                      * source
                      * summary
                      * title
                      * updated_at
                      * url
                      * uuid
                      * related_instruments
                      * preview_text
                      * currency_id

    """ 
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = news_url(symbol)
    data = request_get(url, 'results')

    return(filter_data(data, info))


def get_splits(symbol, info=None):
    """Returns the date, divisor, and multiplier for when a stock split occureed.

    :param symbol: The stock ticker.
    :type symbol: str
    :param info: Will filter the results to get a specific value. Possible options are \
    url, instrument, execution_date, divsor, and multiplier.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries. If info parameter is provided, \
    a list of strings is returned where the strings are the value \
    of the key that matches info.
    :Dictionary Keys: * url
                      * instrument
                      * execution_date
                      * multiplier
                      * divisor

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = splits_url(symbol)
    data = request_get(url, 'results')
    return(filter_data(data, info))


def find_instrument_data(query):
    """Will search for stocks that contain the query keyword and return the instrument data.

    :param query: The keyword to search for.
    :type query: str
    :returns: [list] Returns a list of dictionaries that contain the instrument data for each stock that matches the query.
    :Dictionary Keys: * id
                      * url
                      * quote
                      * fundamentals
                      * splits
                      * state
                      * market
                      * simple_name
                      * name
                      * tradeable
                      * tradability
                      * symbol
                      * bloomberg_unique
                      * margin_initial_ratio
                      * maintenance_ratio
                      * country
                      * day_trade_ratio
                      * list_date
                      * min_tick_size
                      * type
                      * tradable_chain_id
                      * rhs_tradability
                      * fractional_tradability
                      * default_collar_fraction

    """ 
    url = instruments_url()
    payload = {'query': query}

    data = request_get(url, 'pagination', payload)

    if len(data) == 0:
        print('No results found for that keyword', file=get_output())
        return([None])
    else:
        print('Found ' + str(len(data)) + ' results', file=get_output())
        return(data)


def get_stock_historicals(inputSymbols, interval='hour', span='week', bounds='regular', info=None):
    """Represents the historicl data for a stock.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param interval: Interval to retrieve data for. Values are '5minute', '10minute', 'hour', 'day', 'week'. Default is 'hour'.
    :type interval: Optional[str]
    :param span: Sets the range of the data to be either 'day', 'week', 'month', '3month', 'year', or '5year'. Default is 'week'.
    :type span: Optional[str]
    :param bounds: Represents if graph will include extended trading hours or just regular trading hours. Values are 'extended', 'trading', or 'regular'. Default is 'regular'
    :type bounds: Optional[str]
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: [list] Returns a list of dictionaries where each dictionary is for a different time. If multiple stocks are provided \
    the historical data is listed one after another.
    :Dictionary Keys: * begins_at
                      * open_price
                      * close_price
                      * high_price
                      * low_price
                      * volume
                      * session
                      * interpolated
                      * symbol

    """    
    interval_check = ['5minute', '10minute', 'hour', 'day', 'week']
    span_check = ['day', 'week', 'month', '3month', 'year', '5year']
    bounds_check = ['extended', 'regular', 'trading']

    if interval not in interval_check:
        print(
            'ERROR: Interval must be "5minute","10minute","hour","day",or "week"', file=get_output())
        return([None])
    if span not in span_check:
        print('ERROR: Span must be "day","week","month","3month","year",or "5year"', file=get_output())
        return([None])
    if bounds not in bounds_check:
        print('ERROR: Bounds must be "extended","regular",or "trading"', file=get_output())
        return([None])
    if (bounds == 'extended' or bounds == 'trading') and span != 'day':
        print('ERROR: extended and trading bounds can only be used with a span of "day"', file=get_output())
        return([None])

    symbols = inputs_to_set(inputSymbols)
    url = historicals_url()
    payload = {'symbols': ','.join(symbols),
               'interval': interval,
               'span': span,
               'bounds': bounds}

    data = request_get(url, 'results', payload)
    if (data == None or data == [None]):
        return data

    histData = []
    for count, item in enumerate(data):
        if (len(item['historicals']) == 0):
            print(error_ticker_does_not_exist(symbols[count]), file=get_output())
            continue
        stockSymbol = item['symbol']
        for subitem in item['historicals']:
            subitem['symbol'] = stockSymbol
            histData.append(subitem)

    return(filter_data(histData, info))


def get_stock_quote_by_id(stock_id, info=None):
    """
    Represents basic stock quote information

    :param stock_id: robinhood stock id
    :type stock_id: str
    :param info: Will filter the results to get a specific value. Possible options are url, instrument, execution_date, \
    divsor, and multiplier.
    :type info: Optional[str]
    :return: [dict] If the info parameter is provided, then the function will extract the value of the key \
    that matches the info parameter. Otherwise, the whole dictionary is returned.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument
    """
    url = marketdata_quotes_url(stock_id)
    data = request_get(url)

    return (filter_data(data, info))


def get_stock_quote_by_symbol(symbol, info=None):
    """
    Represents basic stock quote information

    :param symbol: robinhood stock id
    :type stock_id: str
    :param info: Will filter the results to get a specific value. Possible options are url, instrument, execution_date, \
    divsor, and multiplier.
    :type info: Optional[str]
    :return: [dict] If the info parameter is provided, then the function will extract the value of the key \
    that matches the info parameter. Otherwise, the whole dictionary is returned.
    :Dictionary Keys: * ask_price
                      * ask_size
                      * bid_price
                      * bid_size
                      * last_trade_price
                      * last_extended_hours_trade_price
                      * previous_close
                      * adjusted_previous_close
                      * previous_close_date
                      * symbol
                      * trading_halted
                      * has_traded
                      * last_trade_price_source
                      * updated_at
                      * instrument
    """

    return get_stock_quote_by_id(id_for_stock(symbol))


def get_pricebook_by_id(stock_id, info=None):
    """
    Represents Level II Market Data provided for Gold subscribers

    :param stock_id: robinhood stock id
    :type stock_id: str
    :param info: Will filter the results to get a specific value. Possible options are url, instrument, execution_date, \
    divsor, and multiplier.
    :type info: Optional[str]
    :return: Returns a dictionary of asks and bids.

    """
    url = marketdata_pricebook_url(stock_id)
    data = request_get(url)

    return (filter_data(data, info))


def get_pricebook_by_symbol(symbol, info=None):
    """
    Represents Level II Market Data provided for Gold subscribers

    :param symbol: symbol id
    :type symbol: str
    :param info: Will filter the results to get a specific value. Possible options are url, instrument, execution_date, \
    divsor, and multiplier.
    :type info: Optional[str]
    :return: Returns a dictionary of asks and bids.

    """

    return get_pricebook_by_id(id_for_stock(symbol))
